# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2022 RERO.
# Copyright (C) 2022 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Database management for Invenio."""

import os
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as package_version
from importlib.resources import files

import sqlalchemy as sa
from flask_alembic import Alembic
from invenio_base.utils import entry_points
from sqlalchemy_utils.functions import get_class_by_table

from .cli import db as db_cmd
from .shared import db
from .utils import versioning_models_registered


class InvenioDB(object):
    """Invenio database extension."""

    def __init__(self, app=None, **kwargs):
        """Extension initialization."""
        self.alembic = Alembic(run_mkdir=False, command_name="alembic")
        if app:
            self.init_app(app, **kwargs)

    def init_app(self, app, **kwargs):
        """Initialize application object."""
        self.init_db(app, **kwargs)

        def pathify(base_entry):
            return str(files(base_entry.module) / os.path.join(base_entry.attr))

        alembic_entry_points = entry_points(group="invenio_db.alembic")
        version_locations = [
            (base_entry.name, pathify(base_entry))
            for base_entry in alembic_entry_points
        ]
        script_location = str(files("invenio_db") / "alembic")
        app.config.setdefault(
            "ALEMBIC",
            {
                "script_location": script_location,
                "version_locations": version_locations,
            },
        )
        app.config.setdefault(
            "ALEMBIC_CONTEXT",
            {
                "transaction_per_migration": True,
                "compare_type": True,  # Allows to detect change of column type, accuracy depends on backend
            },
        )

        self.alembic.init_app(app)
        app.extensions["invenio-db"] = self
        app.cli.add_command(db_cmd)

    def init_db(self, app, entry_point_group="invenio_db.models", **kwargs):
        """Initialize Flask-SQLAlchemy extension."""
        # Setup SQLAlchemy
        app.config.setdefault(
            "SQLALCHEMY_DATABASE_URI",
            "sqlite:///" + os.path.join(app.instance_path, app.name + ".db"),
        )
        app.config.setdefault("SQLALCHEMY_ECHO", False)
        # Needed for before/after_flush/commit/rollback events
        app.config.setdefault("SQLALCHEMY_TRACK_MODIFICATIONS", True)

        # Initialize Flask-SQLAlchemy extension.
        database = kwargs.get("db", db)
        database.init_app(app)

        # Initialize versioning support.
        self.init_versioning(app, database, kwargs.get("versioning_manager"))

        # Initialize model bases
        if entry_point_group:
            for base_entry in entry_points(group=entry_point_group):
                base_entry.load()

        # All models should be loaded by now.
        sa.orm.configure_mappers()
        # Ensure that versioning classes have been built.
        if app.config["DB_VERSIONING"]:
            manager = self.versioning_manager
            if manager.pending_classes:
                if not versioning_models_registered(manager, database.Model):
                    manager.builder.configure_versioned_classes()
            elif "transaction" not in database.metadata.tables:
                manager.declarative_base = database.Model
                manager.create_transaction_model()
                manager.plugins.after_build_tx_class(manager)

    def init_versioning(self, app, database, versioning_manager=None):
        """Initialize the versioning support using SQLAlchemy-Continuum."""
        try:
            package_version("sqlalchemy_continuum")
        except PackageNotFoundError:  # pragma: no cover
            default_versioning = False
        else:
            default_versioning = True

        app.config.setdefault("DB_VERSIONING", default_versioning)

        if not app.config["DB_VERSIONING"]:
            return

        if not default_versioning:  # pragma: no cover
            raise RuntimeError(
                "Please install extra versioning support first by running "
                "pip install invenio-db[versioning]."
            )

        # Now we can import SQLAlchemy-Continuum.
        from sqlalchemy_continuum import make_versioned
        from sqlalchemy_continuum import versioning_manager as default_vm
        from sqlalchemy_continuum.plugins import FlaskPlugin

        # Try to guess user model class:
        if "DB_VERSIONING_USER_MODEL" not in app.config:  # pragma: no cover
            try:
                package_version("invenio_accounts")
            except PackageNotFoundError:
                user_cls = None
            else:
                user_cls = "User"
        else:
            user_cls = app.config.get("DB_VERSIONING_USER_MODEL")

        plugins = [FlaskPlugin()] if user_cls else []

        # Call make_versioned() before your models are defined.
        self.versioning_manager = versioning_manager or default_vm
        make_versioned(
            user_cls=user_cls,
            manager=self.versioning_manager,
            plugins=plugins,
        )

        # Register models that have been loaded beforehand.
        builder = self.versioning_manager.builder

        for tbl in database.metadata.tables.values():
            builder.instrument_versioned_classes(
                database.mapper, get_class_by_table(database.Model, tbl)
            )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Unit of work.

Moved from invenio-records-resources.

Used to group multiple service operations into a single atomic unit. The Unit
of Work maintains a list of operations and coordinates the commit, indexing and
task execution.

Note, this is NOT a clean implementation of the UoW design pattern. The
main purpose of the Unit of Work in Invenio is to coordinate when the database
transaction commit is called, and ensure tasks that have to run after the
transaction are executed (such as indexing and running celery tasks).

This ensures that we can group multiple service calls into a single database
transaction and perform the necessary indexing/task execution afterwards.

**When to use?**

You should use the unit of work instead of running an explicit
``db.session.commit()`` or ``db.session.rollback()`` in the code. Basically
any function where you would normally have called a ``db.session.commit()``
should be changed to something like:

.. code-block:: python

    from invenio_db.uow import \
        ModelCommit, unit_of_work,

    @unit_of_work()
    def create(self, ... , uow=None):
        # ...
        uow.register(ModelCommit(model))
        # ...
        # Do not use `db.session.commit()` in service.

**When not to use?**

If you're not changing the database state there's no need to use the unit of
work. Examples include:

- Reading a model
- Search for models

**How to group multiple service calls?**

In order to group multiple service calls into one atomic operation you can use
the following pattern:

.. code-block:: python

    from invenio_db.uow import UnitOfWork

    with UnitOfWork() as uow:
        # Be careful to always inject "uow" to the service. If not, the
        # service will create its own unit of work and commit.
        service.communities.add(..., uow=uow)
        service.publish(... , uow=uow)
        uow.commit()

If you're not grouping multiple service calls, then simply just call the
service method (and it will commit automatically):

.. code-block:: python

    service.publish(...)

**Writing your own operation?**

You can write your own unit of work operation by subclassing the operation
class and implementing the desired methods:

.. code-block:: python

    from invenio_db.uow import Operation

    class BulkIndexOp(Operation):
        def on_commit(self, uow):
            # ... executed after the database transaction commit ...
"""

from functools import wraps

from .shared import db


#
# Unit of work operations
#
class Operation:
    """Base class for unit of work operations."""

    def on_register(self, uow):
        """Called upon operation registration."""
        pass

    def on_commit(self, uow):
        """Called in the commit phase (after the transaction is committed)."""
        pass

    def on_post_commit(self, uow):
        """Called right after the commit phase."""
        pass

    def on_exception(self, uow, exception):
        """Called in case of an exception."""
        pass

    def on_rollback(self, uow):
        """Called in the rollback phase (after the transaction rollback)."""
        pass

    def on_post_rollback(self, uow):
        """Called right after the rollback phase."""
        pass


class ModelCommitOp(Operation):
    """SQLAlchemy model add/update operation."""

    def __init__(self, model):
        """Initialize the commit operation."""
        super().__init__()
        self._model = model

    def on_register(self, uow):
        """Add model to db session."""
        uow.session.add(self._model)


class ModelDeleteOp(Operation):
    """SQLAlchemy model delete operation."""

    def __init__(self, model):
        """Initialize the set delete operation."""
        super().__init__()
        self._model = model

    def on_register(self, uow):
        """Delete model."""
        uow.session.delete(self._model)


#
# Unit of work context manager
#
class UnitOfWork:
    """Unit of work context manager.

    Note, the unit of work does not currently take duplication of work into
    account. Thus, you can e.g. add two record commit operations of the same
    record which will then index the record twice, even though only one time
    is needed.
    """

    def __init__(self, session=None):
        """Initialize unit of work context."""
        self._session = session or db.session
        self._operations = []
        self._dirty = False

    def __enter__(self):
        """Entering the context."""
        self.session.begin_nested()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Rollback on exception."""
        if exc_type is not None:
            self.rollback(exception=exc_value)
            self._mark_dirty()

    @property
    def session(self):
        """The SQLAlchemy database session associated with this UoW."""
        return self._session

    def _mark_dirty(self):
        """Mark the unit of work as dirty."""
        if self._dirty:
            raise RuntimeError("The unit of work is already committed or rolledback.")
        self._dirty = True

    def commit(self):
        """Commit the unit of work."""
        self.session.commit()
        # Run commit operations
        for op in self._operations:
            op.on_commit(self)
        # Run post commit operations
        for op in self._operations:
            op.on_post_commit(self)
        self._mark_dirty()

    def rollback(self, exception=None):
        """Rollback the database session."""
        self.session.rollback()

        # Run exception operations
        if exception:
            for op in self._operations:
                op.on_exception(self, exception)

            # Commit exception operations
            self.session.commit()

        # Run rollback operations
        for op in self._operations:
            op.on_rollback(self)
        # Run post rollback operations
        for op in self._operations:
            op.on_post_rollback(self)

    def register(self, op):
        """Register an operation."""
        # Run on register
        op.on_register(self)
        # Append to list of operations.
        self._operations.append(op)


def unit_of_work(**kwargs):
    """Decorator to auto-inject a unit of work if not provided.

    If no unit of work is provided, this decorator will create a new unit of
    work and commit it after the function has been executed.

    .. code-block:: python

        @unit_of_work()
        def aservice_method(self, ...., uow=None):
            # ...
            uow.register(...)

    """

    def decorator(f):
        @wraps(f)
        def inner(self, *args, **kwargs):
            if "uow" not in kwargs or kwargs["uow"] is None:
                # Migration path - start a UoW and commit
                with UnitOfWork(db.session) as uow:
                    kwargs["uow"] = uow
                    res = f(self, *args, **kwargs)
                    uow.commit()
                    return res
            else:
                return f(self, *args, **kwargs)

        return inner

    return decorator



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2024 CERN.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Database management for Invenio.

First make sure you have Flask application with Click support (meaning
Flask 0.11+):

    >>> from flask import Flask
    >>> app = Flask('myapp')

Next, initialize your extension:

    >>> from invenio_db import InvenioDB
    >>> db = InvenioDB(app)

Command-line interface
~~~~~~~~~~~~~~~~~~~~~~
Invenio-DB installs the ``db`` command group on your application with the
following commands:

 * ``create`` - Create database tables.
 * ``drop`` - Drop database tables.
 * ``init`` - Initialize database.
 * ``destroy`` - Destroy database.

and ``alembic`` command group for managing upgrade recipes:

 * ``branches`` - Show branch points.
 * ``current`` - Show current revision.
 * ``downgrade`` - Run downgrade migrations.
 * ``heads`` - Show latest revisions.
 * ``log`` - Show revision log.
 * ``merge`` - Create merge revision.
 * ``mkdir`` - Make migration directory.
 * ``revision`` - Create new migration.
 * ``show`` - Show the given revisions.
 * ``stamp`` - Set current revision.
 * ``upgrade`` - Run upgrade migrations.

For more information about how to setup alembic revisions in your module,
please read the section about :doc:`alembic </alembic>`.

Models
~~~~~~

Database models are created by inheriting from the declarative base
``db.Model``:

.. code-block:: python

   # models.py
   from invenio_db import db

   class User(db.Model):
       id = db.Column(db.Integer, primary_key=True)
       username = db.Column(db.String(80), unique=True)
       email = db.Column(db.String(120), unique=True)

Setuptools integration
~~~~~~~~~~~~~~~~~~~~~~

In order for the CLI commands to be aware of your models and upgrade recipies,
you must either import your models in the application factory, or better simply
specify the entry point item in ``invenio_db.models`` group. Invenio-DB then
takes care of loading the models automatically. Alembic configuration of
version locations is assembled from ``invenio_db.alembic`` entry point group.

.. code-block:: python

   # setup.py
   # ...
   setup(
       entry_points={
           'invenio_db.alembic': [
               'branch_name = mymodule:alembic',
           ],
           'invenio_db.models': [
               'mymodule = mymodule.models',
           ],
       },
   )

"""

from .ext import InvenioDB
from .shared import db

__version__ = "2.1.0"

__all__ = (
    "__version__",
    "db",
    "InvenioDB",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
# Copyright (C) 2022 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
# from .signals import secret_key_changed

"""Invenio-DB utility functions."""

from flask import current_app
from sqlalchemy import inspect

from .proxies import current_db
from .shared import db as _db


def rebuild_encrypted_properties(old_key, model, properties, db=_db):
    """Rebuild model's EncryptedType properties when the SECRET_KEY is changed.

    :param old_key: old SECRET_KEY.
    :param model: the affected db model.
    :param properties: list of properties to rebuild.
    """
    inspector = inspect(db.engine)
    primary_key_names = inspector.get_pk_constraint(model.__tablename__)[
        "constrained_columns"
    ]

    new_secret_key = current_app.secret_key
    db.session.expunge_all()
    try:
        with db.session.begin_nested():
            current_app.secret_key = old_key
            db_columns = []
            for primary_key in primary_key_names:
                db_columns.append(getattr(model, primary_key))
            for prop in properties:
                db_columns.append(getattr(model, prop))
            old_rows = db.session.query(*db_columns).all()
    except Exception as e:
        current_app.logger.error(
            "Exception occurred while reading encrypted properties. "
            "Try again before starting the server with the new secret key."
        )
        raise e
    finally:
        current_app.secret_key = new_secret_key
        db.session.expunge_all()

    for old_row in old_rows:
        primary_keys, old_entries = (
            old_row[: len(primary_key_names)],
            old_row[len(primary_key_names) :],
        )
        primary_key_fields = dict(zip(primary_key_names, primary_keys))
        update_values = dict(zip(properties, old_entries))
        model.query.filter_by(**primary_key_fields).update(update_values)
    db.session.commit()


def create_alembic_version_table():
    """Create alembic_version table."""
    alembic = current_app.extensions["invenio-db"].alembic
    if not alembic.migration_context._has_version_table():
        alembic.migration_context._ensure_version_table()
        for head in alembic.script_directory.revision_map._real_heads:
            alembic.migration_context.stamp(alembic.script_directory, head)


def drop_alembic_version_table():
    """Drop alembic_version table."""
    if has_table(current_db.engine, "alembic_version"):
        alembic_version = current_db.Table(
            "alembic_version", current_db.metadata, autoload_with=current_db.engine
        )
        alembic_version.drop(bind=current_db.engine)


def versioning_model_classname(manager, model):
    """Get the name of the versioned model class."""
    if manager.options.get("use_module_name", True):
        return "%s%sVersion" % (
            model.__module__.title().replace(".", ""),
            model.__name__,
        )
    else:
        return "%sVersion" % (model.__name__,)


def versioning_models_registered(manager, base):
    """Return True if all versioning models have been registered."""
    try:
        registry = base.registry._class_registry
    except AttributeError:  # SQLAlchemy <1.4
        registry = base._decl_class_registry
    declared_models = registry.keys()
    return all(
        versioning_model_classname(manager, c) in declared_models
        for c in manager.pending_classes
    )


def alembic_test_context():
    """Alembic test context.

    # skip index from alembic migrations until sqlalchemy 2.0
    # https://github.com/sqlalchemy/sqlalchemy/discussions/7597
    """

    def include_object(object, name, type_, reflected, compare_to):
        if name == "ix_uq_partial_files_object_is_head":
            return False
        return True

    return {
        "transaction_per_migration": True,
        "include_object": include_object,
    }


def has_table(engine, table):
    """Determine if table exists."""
    try:
        return inspect(engine).has_table(table)
    except AttributeError:
        # SQLAlchemy <1.4
        return engine.has_table(table)
